from .utils import load_class_by_path, group_iter
from .context import connections, get_current_config, log, metric
from .redishelpers import redis_zaddbyscore, redis_zpopbyscore, redis_lpopsafe
from .redishelpers import redis_group_command
import time


class Queue(object):

    """ A Queue for Jobs. """

    is_raw = False
    is_timed = False
    is_sorted = False
    is_set = False
    is_reverse = False

    def __init__(self, queue_id):
        if isinstance(queue_id, Queue):
            self.id = queue_id.id  # TODO use __new__?
            self.is_reverse = queue_id.is_reverse
        else:
            if queue_id[-8:] == "_reverse":
                self.is_reverse = True
                queue_id = queue_id[:-8]
            self.id = queue_id

        # Queue types are determined by their suffix.
        if "_raw" in self.id:
            self.is_raw = True

        if "_set" in self.id:
            self.is_set = True
            self.is_raw = True

        if "_timed" in self.id:
            self.is_timed = True
            self.is_sorted = True

        if "_sorted" in self.id:
            self.is_sorted = True

    @property
    def redis_key(self):
        """ Returns the redis key used to store this queue. """
        return "%s:q:%s" % (get_current_config()["redis_prefix"], self.id)

    @classmethod
    def redis_key_started(self):
        """ Returns the global redis key used to store started job ids """
        return "%s:s:started" % get_current_config()["redis_prefix"]

    def get_retry_queue(self):
        """ For raw queues, returns the name of the linked queue for job statuses
            other than "queued" """

        if not self.is_raw:
            return self.id

        return self.get_config().get("retry_queue") or "default"

    def get_config(self):
        """ Returns the specific configuration for this queue """

        return get_current_config().get("raw_queues", {}).get(self.id) or {}

    def size(self):
        """ Returns the total number of jobs on the queue """

        # ZSET
        if self.is_sorted:
            return connections.redis.zcard(self.redis_key)
        # SET
        elif self.is_set:
            return connections.redis.scard(self.redis_key)
        # LIST
        else:
            return connections.redis.llen(self.redis_key)

    def count_jobs_to_dequeue(self):
        """ Returns the number of jobs that can be dequeued right now from the queue. """

        # timed ZSET
        if self.is_timed:
            return connections.redis.zcount(
                self.redis_key,
                "-inf",
                time.time())

        # In all other cases, it's the same as .size()
        else:
            return self.size()

    def list_job_ids(self, skip=0, limit=20):
        """ Returns a list of job ids on a queue """

        if self.is_raw:
            raise Exception("Can't list job ids from a raw queue")

        # ZSET
        if self.is_sorted:
            return connections.redis.zrange(
                self.redis_key,
                skip,
                skip + limit - 1)
        # SET
        elif self.is_set:
            return connections.redis.srandmember(self.redis_key, limit)
        # LIST
        else:
            return connections.redis.lrange(
                self.redis_key,
                skip,
                skip + limit - 1)

    def get_sorted_graph(
            self,
            start=0,
            stop=100,
            slices=100,
            include_inf=False,
            exact=False):
        """ Returns a graph of the distribution of jobs in a sorted set """

        if not self.is_sorted:
            raise Exception("Not a sorted queue")

        with connections.redis.pipeline(transaction=exact) as pipe:
            interval = float(stop - start) / slices
            for i in range(0, slices):
                pipe.zcount(self.redis_key,
                            (start + i * interval),
                            "(%s" % (start + (i + 1) * interval))
            if include_inf:
                pipe.zcount(self.redis_key, stop, "+inf")
                pipe.zcount(self.redis_key, "-inf", "(%s" % start)
            data = pipe.execute()

        if include_inf:
            return data[-1:] + data[:-1]

        return data

    @classmethod
    def all_active(cls):
        """ List active queues, based on their lengths in Redis. """

        prefix = get_current_config()["redis_prefix"]
        queues = []
        for key in connections.redis.keys():
            if key.startswith(prefix):
                queues.append(Queue(key[len(prefix) + 3:]))

        return queues

    @classmethod
    def all(cls):
        """ List all known queues (Raw + MongoDB)"""

        # Start with raw queues we know exist from the config
        queues = {x: 0 for x in get_current_config().get("raw_queues", {})}

        stats = list(connections.mongodb_jobs.mrq_jobs.aggregate([
            {"$match": {"status": "queued"}},
            {"$group": {"_id": "$queue", "jobs": {"$sum": 1}}}
        ])["result"])

        queues.update({x["_id"]: x["jobs"] for x in stats})

        return queues

    def enqueue_job_ids(self, job_ids):
        """ Add Jobs to this queue, once they have been inserted in MongoDB. """

        if len(job_ids) == 0:
            return

        if self.is_raw:
            raise Exception("Can't queue regular jobs on a raw queue")

        # ZSET
        if self.is_sorted:

            if not isinstance(job_ids, dict) and self.is_timed:
                now = time.time()
                job_ids = {str(x): now for x in job_ids}

            connections.redis.zadd(self.redis_key, **job_ids)

        # LIST
        else:
            connections.redis.rpush(self.redis_key, *job_ids)

        metric("queues.%s.enqueued" % self.id, len(job_ids))
        metric("queues.all.enqueued", len(job_ids))

    def enqueue_raw_jobs(self, params_list):
        """ Add Jobs to this queue with raw parameters. They are not yet in MongoDB. """

        if not self.is_raw:
            raise Exception("Can't queue raw jobs in a regular queue")

        if len(params_list) == 0:
            return

        # ZSET
        if self.is_sorted:

            if not isinstance(params_list, dict) and self.is_timed:
                now = time.time()
                params_list = {str(x): now for x in params_list}

            connections.redis.zadd(self.redis_key, **params_list)

        # SET
        elif self.is_set:
            connections.redis.sadd(self.redis_key, *params_list)

        # LIST
        else:
            connections.redis.rpush(self.redis_key, *params_list)

        metric("queues.%s.enqueued" % self.id, len(params_list))
        metric("queues.all.enqueued", len(params_list))

    def remove_raw_jobs(self, params_list):
        """ Remove jobs from a raw queue with their raw params. """

        if not self.is_raw:
            raise Exception("Can't remove raw jobs in a regular queue")

        if len(params_list) == 0:
            return

        # ZSET
        if self.is_sorted:
            connections.redis.zrem(self.redis_key, *iter(params_list))

        # SET
        elif self.is_set:
            connections.redis.srem(self.redis_key, *params_list)

        else:
            # O(n)! Use with caution.
            for k in params_list:
                connections.redis.lrem(self.redis_key, 1, k)

        metric("queues.%s.removed" % self.id, len(params_list))
        metric("queues.all.removed", len(params_list))

    def empty(self):
        """ Empty a queue. """
        return connections.redis.delete(self.redis_key)

    def dequeue_jobs(self, max_jobs=1, job_class=None, worker=None):
        """ Fetch a maximum of max_jobs from this queue """

        if job_class is None:
            from .job import Job
            job_class = Job

        # Used in tests to simulate workers exiting abruptly
        simulate_zombie_jobs = get_current_config().get("simulate_zombie_jobs")

        jobs = []

        if self.is_raw:

            queue_config = self.get_config()

            job_factory = queue_config.get("job_factory")
            if not job_factory:
                raise Exception("No job_factory configured for raw queue %s" % self.id)

            retry_queue = self.get_retry_queue()

            params = []

            # ZSET with times
            if self.is_timed:

                current_time = time.time()

                # When we have a pushback_seconds argument, we never pop items from the queue, instead
                # we push them back by an amount of time so that they don't get dequeued again until
                # the task finishes.

                # TODO add reverse queue support

                pushback_time = current_time + float(queue_config.get("pushback_seconds") or 0)
                if pushback_time > current_time:
                    params = redis_zaddbyscore()(
                        keys=[self.redis_key],
                        args=[
                            "-inf", current_time, 0, max_jobs, pushback_time
                        ])

                else:
                    params = redis_zpopbyscore()(
                        keys=[self.redis_key],
                        args=[
                            "-inf", current_time, 0, max_jobs
                        ])

            # ZSET
            elif self.is_sorted:

                # TODO Lua?
                with connections.redis.pipeline(transaction=True) as pipe:
                    pipe.zrange(self.redis_key, 0, max_jobs - 1)
                    pipe.zremrangebyrank(self.redis_key, 0, max_jobs - 1)
                    params = pipe.execute()[0]

            # SET
            elif self.is_set:
                params = redis_group_command("spop", max_jobs, self.redis_key)

            # LIST
            else:
                params = redis_group_command("lpop", max_jobs, self.redis_key)

            if len(params) == 0:
                return []

            # Caution, not having a pushback_time may result in lost jobs if the worker interrupts
            # before the mongo insert!
            if simulate_zombie_jobs:
                return []

            if worker:
                worker.status = "spawn"

            job_data = [job_factory(p) for p in params]
            for j in job_data:
                j["status"] = "started"
                j["queue"] = retry_queue
                if worker:
                    j["worker"] = worker.id

            jobs += job_class.insert(job_data)

        # Regular queue, in a LIST
        else:

            # TODO implement _timed and _sorted queues here.

            job_ids = redis_lpopsafe()(keys=[
                self.redis_key,
                Queue.redis_key_started()
            ], args=[
                max_jobs,
                time.time(),
                "0" if self.is_reverse else "1"
            ])

            if len(job_ids) == 0:
                return []

            # At this point, the job is in the redis started zset but not in Mongo yet.
            # It may become "zombie" if we interrupt here but we can recover it from
            # the started zset.
            if simulate_zombie_jobs:
                return []

            if worker:
                worker.status = "spawn"

            jobs += [job_class(_job_id, queue=self.id, start=True)
                     for _job_id in job_ids if _job_id]

            # Now that the jobs have been marked as started in Mongo, we can
            # remove them from the started queue.
            connections.redis.zrem(Queue.redis_key_started(), *job_ids)

        for job in jobs:
            metric("queues.%s.dequeued" % job.queue, 1)
        metric("queues.all.dequeued", len(jobs))

        return jobs


def send_raw_tasks(queue, params_list, **kwargs):
    """ Queue some tasks on a raw queue """

    q = Queue(queue)
    q.enqueue_raw_jobs(params_list, **kwargs)


def send_task(path, params, **kwargs):
    """ Queue one task on a regular queue """

    return send_tasks(path, [params], **kwargs)[0]


def send_tasks(path, params_list, queue=None, sync=False, batch_size=1000):
    """ Queue several task on a regular queue """

    if len(params_list) == 0:
        return []

    if sync:
        task_class = load_class_by_path(path)
        return [task_class().run(params) for params in params_list]

    if queue is None:
        task_def = get_current_config().get("tasks", {}).get(path) or {}
        queue = task_def.get("queue", "default")

    queue_obj = Queue(queue)

    if queue_obj.is_raw:
        raise Exception("Can't queue regular jobs on a raw queue")

    all_ids = []

    # Avoid circular import
    from .job import Job

    for params_group in group_iter(params_list, n=batch_size):

        metric("jobs.status.queued", len(params_group))

        job_ids = Job.insert([{
            "path": path,
            "params": params,
            "queue": queue,
            "status": "queued"
        } for params in params_group], w=1, return_jobs=False)

        # Between these 2 calls, a task can be inserted in MongoDB but not queued in Redis.
        # This is the same as dequeueing a task from Redis and being stopped before updating the "started"
        # flag in MongoDB.

        queue_obj.enqueue_job_ids([str(x) for x in job_ids])

        all_ids += job_ids

    return all_ids



import argparse
import os
import sys
from .version import VERSION
from .utils import get_local_ip
import atexit


def add_parser_args(parser, config_type):

    # General arguments

    parser.add_argument(
        '--trace_greenlets',
        action='store_true',
        default=False,
        help='Collect stats about each greenlet execution time and switches.')

    parser.add_argument(
        '--trace_memory',
        action='store_true',
        default=False,
        help='Collect stats about memory for each task. Incompatible with gevent > 1')

    parser.add_argument(
        '--trace_io',
        action='store_true',
        default=False,
        help='Collect stats about all I/O operations')

    parser.add_argument(
        '--print_mongodb',
        action='store_true',
        default=False,
        help='Print all MongoDB requests')

    parser.add_argument(
        '--trace_memory_type',
        action='store',
        default="",
        help='Create a .png object graph in trace_memory_output_dir with a random object of this type.')

    parser.add_argument(
        '--trace_memory_output_dir',
        action='store',
        default="memory_traces",
        help='Directory where to output .pngs with object graphs')

    parser.add_argument(
        '--profile',
        action='store_true',
        default=False,
        help='Run profiling on the whole worker')

    parser.add_argument(
        '--mongodb_jobs',
        action='store',
        default="mongodb://127.0.0.1:27017/mrq",
        help='MongoDB URI for the jobs, scheduled_jobs & workers database')

    parser.add_argument(
        '--mongodb_logs',
        action='store',
        default="mongodb://127.0.0.1:27017/mrq",
        help='MongoDB URI for the logs database. If set to "0", will disable remote logs.')

    parser.add_argument(
        '--mongodb_logs_size',
        action='store',
        default=16 *
        1024 *
        1024,
        type=int,
        help='If provided, sets the log collection to capped to that amount of bytes.')

    parser.add_argument(
        '--no_mongodb_ensure_indexes',
        action='store_true',
        default=False,
        help='If provided, skip the creation of MongoDB indexes at worker startup.')

    parser.add_argument(
        '--redis',
        action='store',
        default="redis://127.0.0.1:6379",
        help='Redis URI')

    parser.add_argument(
        '--redis_prefix',
        action='store',
        default="mrq",
        help='Redis key prefix')

    parser.add_argument(
        '--name',
        default=None,
        action='store',
        help='Specify a different name')

    parser.add_argument(
        '--quiet',
        default=False,
        action='store_true',
        help='Don\'t output task logs')

    parser.add_argument(
        '--config',
        '-c',
        default=None,
        action="store",
        help='Path of a config file')

    parser.add_argument(
        '--worker_class',
        default="mrq.worker.Worker",
        action="store",
        help='Path to a custom worker class')

    parser.add_argument(
        '--version',
        '-v',
        default=False,
        action="store_true",
        help='Prints current MRQ version')

    parser.add_argument(
        '--no_import_patch',
        default=False,
        action='store_true',
        help='Skips patching __import__ to fix gevent bug #108')

    parser.add_argument(
        '--add_network_latency',
        default="0",
        action='store',
        type=str,
        help='Adds random latency to the network calls, zero to N seconds. Can be a range (1-2)')

    # mrq-run-specific arguments

    if config_type == "run":

        parser.add_argument(
            '--async',
            action='store_true',
            default=False,
            help='Queue the task instead of running it right away')

        parser.add_argument(
            '--queue',
            action='store',
            default="default",
            help='Queue where to put the task when async')

        parser.add_argument(
            'taskpath',
            action='store',
            help='Task to run')

        parser.add_argument(
            'taskargs',
            action='store',
            default='{}',
            nargs='*',
            help='JSON-encoded arguments, or "key value" pairs')

    # Dashboard-specific arguments

    elif config_type == "dashboard":

        parser.add_argument(
            '--dashboard_httpauth',
            default="",
            action="store",
            help='HTTP Auth for the Dashboard. Format is user:pass')

        parser.add_argument(
            '--dashboard_queue',
            default=None,
            action="store",
            help='Default queue for dashboard actions.')

    # Worker-specific args

    elif config_type == "worker":

        parser.add_argument(
            '--max_jobs',
            default=0,
            type=int,
            action='store',
            help='Gevent: max number of jobs to do before quitting. Temp workaround for memory leaks')

        parser.add_argument(
            '--gevent',
            '-g',
            default=1,
            type=int,
            action='store',
            help='Gevent: max number of greenlets')

        parser.add_argument(
            '--processes',
            '-p',
            default=0,
            type=int,
            action='store',
            help='Number of processes to launch with supervisord')

        default_template = os.path.abspath(os.path.join(
            os.path.dirname(__file__),
            "supervisord_templates/default.conf"
        ))

        parser.add_argument(
            '--supervisord_template',
            default=default_template,
            action='store',
            help='Path of supervisord template to use')

        parser.add_argument(
            '--scheduler',
            default=False,
            action='store_true',
            help='Run the scheduler')

        parser.add_argument(
            '--scheduler_interval',
            default=60,
            action='store',
            type=int,
            help='Seconds between scheduler checks')

        parser.add_argument(
            '--report_interval',
            default=10,
            action='store',
            type=float,
            help='Seconds between worker reports to MongoDB')

        parser.add_argument(
            '--report_file',
            default="",
            action='store',
            type=unicode,
            help='Filepath of a json dump of the worker status. Disabled if none')

        parser.add_argument(
            'queues',
            nargs='*',
            default=["default"],
            help='The queues to listen on (default: \'default\')')

        parser.add_argument(
            '--admin_port',
            default=0,
            action="store",
            type=int,
            help='Start an admin server on this port, if provided. Incompatible with --processes')

        parser.add_argument(
            '--local_ip',
            default=get_local_ip(),
            action="store",
            type=str,
            help='Overwrite the local IP, to be displayed in the dashboard.')


def get_config(
        sources=(
            "file",
            "env",
            "args"),
        env_prefix="MRQ_",
        file_path=None,
        parser=None,
        config_type="worker"):
    """ Returns a config dict merged from several possible sources """

    if not parser:
        parser = argparse.ArgumentParser()

    add_parser_args(parser, config_type)

    if config_type in ["run"]:
        default_config = parser.parse_args(["x"]).__dict__
    else:
        default_config = parser.parse_args([]).__dict__

    # Keys that can't be passed from the command line
    default_config["tasks"] = {}
    default_config["scheduled_tasks"] = {}

    # Only keep values different from config, actually passed on the command
    # line
    from_args = {}
    if "args" in sources:
        for k, v in parser.parse_args().__dict__.iteritems():
            if default_config[k] != v:
                from_args[k] = v

    # If we were given another config file, use it
    if file_path is not None:
        config_file = file_path
    elif from_args.get("config"):
        config_file = from_args.get("config")
    # If a mrq-config.py file is in the current directory, use it!
    elif os.path.isfile(os.path.join(os.getcwd(), "mrq-config.py")):
        config_file = os.path.join(os.getcwd(), "mrq-config.py")
    else:
        config_file = None

    from_file = {}
    if config_file and "file" in sources:
        sys.path.insert(0, os.path.dirname(config_file))
        config_module = __import__(os.path.basename(config_file.replace(".py", "")))
        sys.path.pop(0)
        for k, v in config_module.__dict__.iteritems():

            # We only keep variables starting with an uppercase character.
            if k[0].isupper():
                default_config[k.lower()] = v
                if k.lower() not in default_config:
                    default_config[k.lower()] = v

    # Merge the config in the order given by the user
    merged_config = default_config
    for part in sources:
        for name in merged_config:

            if part == "env":
                value = os.environ.get(env_prefix + name.upper())
                if value:
                    merged_config[name] = value
            elif part == "args" and name in from_args:
                merged_config[name] = from_args[name]
            elif part == "file" and name in from_file:
                merged_config[name] = from_file[name]

    if merged_config["profile"]:
        import cProfile
        profiler = cProfile.Profile()
        profiler.enable()

        def print_profiling():
            profiler.print_stats(sort="cumulative")

        atexit.register(print_profiling)

    if merged_config["version"]:
        print "MRQ version: %s" % VERSION
        print "Python version: %s" % sys.version
        sys.exit(1)

    return merged_config



import datetime
from bson import ObjectId
import pymongo
import time
from .exceptions import RetryInterrupt, CancelInterrupt
from .utils import load_class_by_path
from .queue import Queue
from .context import get_current_worker, log, connections, get_current_config, set_current_job, metric
import gevent
import objgraph
import random
import gc
from collections import defaultdict
import traceback
import sys


class Job(object):

    # Seconds the job can last before timeouting
    timeout = 300

    # Seconds the results are kept in MongoDB
    result_ttl = 7 * 24 * 3600

    progress = 0

    # Exceptions that don't mark the task as failed but as retry
    retry_on_exceptions = (
        pymongo.errors.AutoReconnect,
        pymongo.errors.OperationFailure,
        pymongo.errors.ConnectionFailure,
        RetryInterrupt
    )

    # Exceptions that will make the task as cancelled
    cancel_on_exceptions = (
        CancelInterrupt
    )

    _memory_stop = 0
    _memory_start = 0

    _current_io = None

    def __init__(self, job_id, queue=None, start=False, fetch=False):
        self.worker = get_current_worker()
        self.queue = queue
        self.datestarted = None

        self.collection = connections.mongodb_jobs.mrq_jobs

        if job_id is None:
            self.id = None
        else:
            self.id = ObjectId(job_id)

        self.data = None
        self.saved = True

        self.task = None
        self.greenlet_switches = 0
        self.greenlet_time = 0

        self._trace_mongodb = defaultdict(int)

        if start:
            self.fetch(start=True, full_data=False)
        elif fetch:
            self.fetch(start=False, full_data=False)

    def exists(self):
        return bool(self.collection.find_one({"_id": self.id}, fields={"_id": 1}))

    def fetch(self, start=False, full_data=True):
        """ Get the current job data and possibly flag it as started. """

        if self.id is None:
            return self

        if full_data is True:
            fields = None
        elif isinstance(full_data, dict):
            fields = full_data
        else:
            fields = {
                "_id": 0,
                "path": 1,
                "params": 1,
                "status": 1
            }

        if start:

            self.datestarted = datetime.datetime.utcnow()
            self.set_data(self.collection.find_and_modify(
                {
                    "_id": self.id,
                    "status": {"$nin": ["cancel"]}
                },
                {"$set": {
                    "status": "started",
                    "datestarted": self.datestarted,
                    "worker": self.worker.id
                }},
                fields=fields)
            )

            metric("jobs.status.started")

        else:
            self.set_data(self.collection.find_one({
                "_id": self.id
            }, fields=fields))

        if self.data is None:
            log.info(
                "Job %s not found in MongoDB or status was cancelled!" %
                self.id)

        return self

    def set_data(self, data):
        self.data = data
        if self.data is None:
            return

        if "path" in self.data:
            task_def = get_current_config().get("tasks", {}).get(
                self.data["path"]
            ) or {}
            self.timeout = task_def.get("timeout", self.timeout)
            self.result_ttl = task_def.get("result_ttl", self.result_ttl)

    def set_progress(self, ratio, save=False):
        self.data["progress"] = ratio
        self.saved = False

        # If not saved, will be updated in the next worker report
        if save:
            self.save()

    def save(self):
        """ Will be called at each worker report. """

        if not self.saved and self.data and "progress" in self.data:
            # TODO should we save more fields?
            self.collection.update({"_id": self.id}, {"$set": {
                "progress": self.data["progress"]
            }})
            self.saved = True

    @classmethod
    def insert(self, jobs_data, queue=None, return_jobs=True, w=1):
        """ Insert a job into MongoDB """

        now = datetime.datetime.utcnow()
        for data in jobs_data:
            if data["status"] == "started":
                data["datestarted"] = now

        inserted = connections.mongodb_jobs.mrq_jobs.insert(
            jobs_data,
            manipulate=True,
            w=w
        )

        if return_jobs:
            jobs = []
            for data in jobs_data:
                job = self(data["_id"], queue=queue)
                job.set_data(data)
                if data["status"] == "started":
                    job.datestarted = data["datestarted"]
                jobs.append(job)

            return jobs
        else:
            return inserted

    def save_status(
            self,
            status,
            result=None,
            exception=False,
            dateretry=None,
            queue=None,
            w=1):

        if self.id is None:
            return

        now = datetime.datetime.utcnow()
        updates = {
            "status": status,
            "dateupdated": now
        }

        if self.datestarted:
            updates["totaltime"] = (now - self.datestarted).total_seconds()
        if result is not None:
            updates["result"] = result

        if dateretry is not None:
            updates["dateretry"] = dateretry
        if queue is not None:
            updates["queue"] = queue
        if get_current_config().get("trace_greenlets"):
            current_greenlet = gevent.getcurrent()

            # TODO are we sure the current job is doing the save_status() on itself?
            if hasattr(current_greenlet, "_trace_time"):
                updates["time"] = current_greenlet._trace_time
                updates["switches"] = current_greenlet._trace_switches

        if exception:
            trace = traceback.format_exc()
            log.error(trace)
            updates["traceback"] = trace
            updates["exceptiontype"] = sys.exc_info()[0].__name__

        # Make the job document expire
        if status in ("success", "cancel"):
            updates["dateexpires"] = now + \
                datetime.timedelta(seconds=self.result_ttl)

        if status == "success" and "progress" in self.data:
            updates["progress"] = 1

        self.collection.update({
            "_id": self.id
        }, {"$set": updates}, w=w, manipulate=False)

        metric("jobs.status.%s" % status)

        if self.data:
            self.data.update(updates)

    def save_retry(self, exc, exception=False):

        countdown = 24 * 3600

        if isinstance(exc, RetryInterrupt) and exc.countdown is not None:
            countdown = exc.countdown

        queue = None
        if isinstance(exc, RetryInterrupt) and exc.queue:
            queue = exc.queue

        if countdown == 0:
            self.requeue(queue=queue)
        else:
            now = datetime.datetime.utcnow()
            self.save_status(
                "retry",
                exception=exception,
                dateretry=now + datetime.timedelta(seconds=countdown),
                queue=queue
            )

    def retry(self, queue=None, countdown=None, max_retries=None):

        if self.task.cancel_on_retry:
            raise CancelInterrupt()
        else:
            exc = RetryInterrupt()
            exc.queue = queue
            exc.countdown = countdown
            raise exc

    def cancel(self):
        self.save_status("cancel")

    def requeue(self, queue=None):

        if not queue:
            if not self.data or not self.data.get("queue"):
                self.fetch(full_data={"_id": 0, "queue": 1, "path": 1})
            queue = self.data["queue"]

        queue_obj = Queue(queue)

        self.save_status("queued", queue=queue)

        # Between these two lines, jobs can become "lost" too.

        queue_obj.enqueue_job_ids([str(self.id)])

    def perform(self):
        """ Loads and starts the main task for this job, the saves the result. """

        if self.data is None:
            return

        log.debug("Starting %s(%s)" % (self.data["path"], self.data["params"]))
        task_class = load_class_by_path(self.data["path"])

        self.task = task_class()

        self.task.is_main_task = True

        result = self.task.run(self.data["params"])

        self.save_status("success", result)

        if get_current_config().get("trace_greenlets"):

            # TODO: this is not the exact greenlet_time measurement because it doesn't
            # take into account the last switch's time. This is why we force a last switch.
            # This does cause a performance overhead. Instead, we should print the
            # last timing directly from the trace() function in context?

            gevent.sleep(0)
            current_greenlet = gevent.getcurrent()
            t = (datetime.datetime.utcnow() - self.datestarted).total_seconds()
            log.debug(
                "Job %s success: %0.6fs total, %0.6fs in greenlet, %s switches" %
                (self.id,
                 t,
                 current_greenlet._trace_time,
                 current_greenlet._trace_switches - 1)
            )

        else:
            log.debug("Job %s success: %0.6fs total" % (
                self.id, (datetime.datetime.utcnow() -
                          self.datestarted).total_seconds()
            ))

        return result

    def wait(self, poll_interval=1, timeout=None, full_data=False):
        """ Wait for this job to finish. """

        end_time = None
        if timeout:
            end_time = time.time() + timeout

        while (end_time is None or time.time() < end_time):

            job_data = self.collection.find_one({
                "_id": ObjectId(self.id),
                "status": {"$nin": ["started", "queued"]}
            }, fields=({
                "_id": 0,
                "result": 1,
                "status": 1
            } if not full_data else None))

            if job_data:
                return job_data

            time.sleep(poll_interval)

        raise Exception(
            "Waited for job result for %ss seconds, timeout." % timeout)

    def set_current_io(self, io_data):
        if io_data is None:
            if not self._current_io:
                return

            t = time.time() - self._current_io["started"]
            if self.worker:
                self.worker._traced_io["types"][self._current_io["type"]] += t
                self.worker._traced_io["tasks"][self.data["path"]] += t
                self.worker._traced_io["total"] += t

            self._current_io = None

        else:
            io_data["started"] = time.time()
            self._current_io = io_data

    def trace_memory_start(self):
        """ Starts measuring memory consumption """

        objgraph.show_growth(limit=10)

        gc.collect()
        self._memory_start = self.worker.get_memory()

    def subpool_map(self, *args, **kwargs):
        """ Deprecated! Use from mrq.context """
        from mrq.context import subpool_map
        return subpool_map(*args, **kwargs)

    def trace_memory_stop(self):
        """ Stops measuring memory consumption """

        objgraph.show_growth(limit=10)

        trace_type = get_current_config()["trace_memory_type"]
        if trace_type:

            filename = '%s/%s-%s.png' % (
                get_current_config()["trace_memory_output_dir"],
                trace_type,
                self.id)

            objgraph.show_chain(
                objgraph.find_backref_chain(
                    random.choice(
                        objgraph.by_type(trace_type)
                    ),
                    objgraph.is_proper_module
                ),
                filename=filename
            )

        gc.collect()
        self._memory_stop = self.worker.get_memory()

        diff = self._memory_stop - self._memory_start

        log.debug("Memory diff for job %s : %s" % (self.id, diff))

        # We need to update it later than the results, we need them off memory
        # already.
        self.collection.update(
            {"_id": self.id},
            {"$set": {
                "memory_diff": diff
            }},
            w=1
        )



from mrq.queue import send_task, Queue
from mrq.task import Task
from mrq.job import Job
from mrq.context import log, connections
import datetime
import time


class RequeueInterruptedJobs(Task):

    """ Requeue jobs that were marked as status=interrupt when a worker got a SIGTERM. """

    def run(self, params):
        return send_task("mrq.basetasks.utils.JobAction", {
            "status": "interrupt",
            "action": "requeue"
        }, sync=True)


class RequeueRetryJobs(Task):

    """ Requeue jobs that were marked as retry. """

    def run(self, params):
        return send_task("mrq.basetasks.utils.JobAction", {
            "status": "retry",
            "dateretry": {"$lte": datetime.datetime.utcnow()},
            "action": "requeue"
        }, sync=True)


class RequeueStartedJobs(Task):

    """ Requeue jobs that were marked as status=started and never finished.

        That may be because the worker got a SIGKILL or was terminated abruptly. The timeout parameter
        of this task is in addition to the task's own timeout.
    """

    def run(self, params):

        additional_timeout = params.get("timeout", 300)

        stats = {
            "requeued": 0,
            "started": 0
        }

        # There shouldn't be that much "started" jobs so we can quite safely
        # iterate over them.
        self.collection = connections.mongodb_jobs.mrq_jobs
        for job_data in self.collection.find(
                {"status": "started"}, fields={"_id": 1, "datestarted": 1, "queue": 1, "path": 1}):
            job = Job(job_data["_id"])
            job.set_data(job_data)

            stats["started"] += 1

            expire_date = datetime.datetime.utcnow(
            ) - datetime.timedelta(seconds=job.timeout + additional_timeout)

            if job_data["datestarted"] < expire_date:
                log.debug("Requeueing job %s" % job.id)
                job.requeue()
                stats["requeued"] += 1

        return stats


class RequeueRedisStartedJobs(Task):

    """ Requeue jobs that were started in Redis but not in Mongo.

        They could have been lost by a worker interrupt between
        redis.lpop and mongodb.update
    """

    def run(self, params):

        self.collection = connections.mongodb_jobs.mrq_jobs

        redis_key_started = Queue.redis_key_started()

        stats = {
            "fetched": 0,
            "requeued": 0
        }

        # Fetch all the jobs started more than a minute ago - they should not
        # be in redis:started anymore
        job_ids = connections.redis.zrangebyscore(
            redis_key_started, "-inf", time.time() - params.get("timeout", 60))

        for job_id in job_ids:

            queue = Job(job_id, start=False, fetch=False).fetch(
                full_data=True).data["queue"]

            stats["fetched"] += 1

            log.info("Requeueing %s on %s" % (job_id, queue))

            # TODO LUA script & don't rpush if not in zset anymore.
            with connections.redis.pipeline(transaction=True) as pipeline:
                pipeline.zrem(redis_key_started, job_id)
                pipeline.rpush(Queue(queue).redis_key, job_id)
                pipeline.execute()

            stats["requeued"] += 1

        return stats


class RequeueLostJobs(Task):

    """ Requeue jobs that were queued but don't appear in Redis anymore.

        They could have been lost either by a Redis flush
    """

    def run(self, params):

        self.collection = connections.mongodb_jobs.mrq_jobs

        # If there are more than this much items on the queue, we don't try to check if our mongodb
        # jobs are still queued.
        max_queue_items = params.get("max_queue_items", 1000)

        stats = {
            "fetched": 0,
            "requeued": 0
        }

        for job_data in self.collection.find({
            "status": "queued"
        }, fields={"_id": 1, "queue": 1}).sort([("_id", 1)]):

            stats["fetched"] += 1

            queue = Queue(job_data["queue"])
            queue_size = queue.size()
            if queue_size > max_queue_items:
                log.info("Stopping because queue %s has %s items" %
                         (queue, queue_size))
                break

            queue_jobs_ids = set(queue.list_job_ids(limit=max_queue_items + 1))
            if len(queue_jobs_ids) >= max_queue_items:
                log.info(
                    "Stopping because queue %s actually had more than %s items" %
                    (queue, len(queue_jobs_ids)))
                break

            if str(job_data["_id"]) in queue_jobs_ids:
                log.info("Stopping because we found job %s in redis" %
                         job_data["_id"])
                break

            # At this point, this job is not on the queue and we're sure the queue is less than max_queue_items
            # We can safely requeue the job.
            log.info("Requeueing %s on %s" % (job_data["_id"], queue.id))

            stats["requeued"] += 1
            job = Job(job_data["_id"])
            job.requeue(queue=job_data["queue"])

        return stats


# class CleanupStoppedWorkers(Task):
#   """ Remove stopped or timeouted workers from mongodb.mrq_workers. """

#   def run(self, params):



from .logger import LoggerInterface
import gevent
import gevent.pool
import urlparse
import re
import time
from .utils import LazyObject
from itertools import count as itertools_count


# greenletid => Job object
_GREENLET_JOBS_REGISTRY = {}

_WORKER = None

_CONFIG = None

# Global log object, usable from all tasks
log = LoggerInterface(None, job="current")


def set_current_job(job):
    current = gevent.getcurrent()

    current.__dict__["_trace_time"] = 0
    current.__dict__["_trace_switches"] = 0

    if job is None:
        if id(current) in _GREENLET_JOBS_REGISTRY:
            del _GREENLET_JOBS_REGISTRY[id(current)]
    else:
        _GREENLET_JOBS_REGISTRY[id(current)] = job


def get_current_job(greenlet_id=None):
    if greenlet_id is None:
        greenlet_id = id(gevent.getcurrent())
    return _GREENLET_JOBS_REGISTRY.get(greenlet_id)


def set_current_worker(worker):
    global _WORKER
    _WORKER = worker
    set_current_config(worker.config)


def get_current_worker():
    global _WORKER
    return _WORKER


def set_current_config(config):
    global _CONFIG
    _CONFIG = config
    log.quiet = config["quiet"]

    if config["add_network_latency"] != "0" and config["add_network_latency"]:
        from mrq.monkey import patch_network_latency
        patch_network_latency(config["add_network_latency"])

    if config["print_mongodb"] or config["trace_io"]:
        from mrq.monkey import patch_pymongo
        patch_pymongo(config)

    if config["trace_io"]:
        from mrq.monkey import patch_io_all
        patch_io_all()

    if not config["no_import_patch"]:
        from mrq.monkey import patch_import
        patch_import()


def get_current_config():
    global _CONFIG
    return _CONFIG


def retry_current_job(**kwargs):
    get_current_job().retry(**kwargs)


def _connections_factory(attr):

    config = get_current_config()

    # Connection strings may be stored directly in config
    config_obj = config.get(attr)

    if attr.startswith("redis"):
        if type(config_obj) in [str, unicode]:

            import redis as pyredis

            urlparse.uses_netloc.append('redis')
            redis_url = urlparse.urlparse(config_obj)

            log.info("%s: Connecting to Redis at %s..." %
                     (attr, redis_url.hostname))

            redis_pool = pyredis.ConnectionPool(
                host=redis_url.hostname,
                port=redis_url.port,
                db=int((redis_url.path or "").replace("/", "") or "0"),
                password=redis_url.password
            )
            return pyredis.StrictRedis(connection_pool=redis_pool)

        # Let's just assume we got a StrictRedis-like object!
        else:
            return config_obj

    elif attr.startswith("mongodb"):

        if type(config_obj) in [str, unicode]:

            try:
                # MongoKit's Connection object is just a wrapped MongoClient.
                from mongokit import Connection as MongoClient
                from mongokit import ReplicaSetConnection as MongoReplicaSetClient
            except ImportError:
                from pymongo import MongoClient
                from pymongo import MongoReplicaSetClient

            (
                mongoAuth,
                mongoUsername,
                mongoPassword,
                mongoHosts,
                mongoDbName,
                mongoDbOptions
            ) = re.match(
                r"mongodb://((\w+):(\w+)@)?([\w\.:,-]+)/([\w-]+)(\?.*)?",
                config_obj
            ).groups()

            log.debug("%s: Connecting to MongoDB at %s/%s..." %
                      (attr, mongoHosts, mongoDbName))

            kwargs = {"use_greenlets": True}
            options = {}
            if mongoDbOptions:
                options = {
                    k: v[0]
                    for k, v in urlparse.parse_qs(mongoDbOptions[1:]).iteritems()
                }

            # We automatically switch to MongoReplicaSetClient when getting a replicaSet option.
            # This should cover most use-cases.
            # http://api.mongodb.org/python/current/examples/high_availability.html#mongoreplicasetclient
            if options.get("replicaSet"):
                db = MongoReplicaSetClient(config_obj, **kwargs)[mongoDbName]
            else:
                db = MongoClient(config_obj, **kwargs)[mongoDbName]
            log.debug("%s: ... connected." % (attr))

            return db

        # Let's just assume we got a MongoDB-like object!
        else:
            return config_obj

connections = LazyObject()
connections.add_factory(_connections_factory)
del _connections_factory


def enable_greenlet_tracing():

    # Tracing seems to cause a 2-5% performance loss.

    import greenlet
    greenlet.GREENLET_USE_TRACING = True

    def trace(*args):

        time_since_last_switch = time.time() - trace.last_switch

        # Record the time of the current switch
        trace.last_switch = time.time()

        if args[0] == "switch":
            # We are switching from the greenlet args[1][0] to the greenlet
            # args[1][1]
            args[1][0].__dict__.setdefault("_trace_time", 0)
            args[1][0].__dict__["_trace_time"] += time_since_last_switch
            args[1][0].__dict__.setdefault("_trace_switches", 0)
            args[1][0].__dict__["_trace_switches"] += 1

        elif args[0] == "throw":
            pass

    trace.last_switch = time.time()

    greenlet.settrace(trace)  # pylint: disable-msg=E1101


def progress(ratio, save=False):
    job = get_current_job()
    if not job:
        return
    job.set_progress(ratio, save=save)


def metric(name, incr=1, **kwargs):
    cfg = get_current_config()
    if cfg.get("metric_hook"):
        return cfg.get("metric_hook")(name, incr=incr, **kwargs)


def subpool_map(pool_size, func, iterable):
    """ Starts a Gevent pool and run a map. Takes care of setting current_job and cleaning up. """

    if not pool_size:
        return [func(*args) for args in iterable]

    counter = itertools_count()

    current_job = get_current_job()

    def inner_func(*args):
        next(counter)
        if current_job:
            set_current_job(current_job)
        ret = func(*args)
        if current_job:
            set_current_job(None)
        return ret

    start_time = time.time()
    pool = gevent.pool.Pool(size=pool_size)
    ret = pool.map(inner_func, iterable)
    pool.join(raise_error=True)
    total_time = time.time() - start_time

    log.debug("SubPool ran %s greenlets in %0.6fs" % (counter, total_time))

    return ret



from .context import log
from .queue import send_task
import datetime


class Scheduler(object):

    def __init__(self, collection):
        self.collection = collection

        self.refresh()

    def refresh(self):
        self.all_tasks = list(self.collection.find())

    def hash_task(self, task):
        return " ".join(
            [str(task.get(x)) for x in ["path", "params", "interval", "dailytime", "queue"]])

    def sync_tasks(self, tasks):
        """ Performs the first sync of a list of tasks, often defined in the config file. """

        tasks_by_hash = {self.hash_task(t): t for t in tasks}

        for task in self.all_tasks:
            if tasks_by_hash.get(task["hash"]):
                del tasks_by_hash[task["hash"]]
            else:
                self.collection.remove({"_id": task["_id"]})
                log.debug("Scheduler: deleted %s" % task["hash"])

        for h, task in tasks_by_hash.iteritems():
            task["hash"] = h
            task["datelastqueued"] = datetime.datetime.fromtimestamp(0)
            if task.get("dailytime"):
                # Because MongoDB can store datetimes but not times, we add today's date to the dailytime.
                # The date part will be discarded in check()
                task["dailytime"] = datetime.datetime.combine(
                    datetime.datetime.utcnow(), task["dailytime"])
                task["interval"] = 3600 * 24

            self.collection.insert(task)
            log.debug("Scheduler: added %s" % task["hash"])

        self.refresh()

    def check(self):

        log.debug(
            "Scheduler checking for out-of-date scheduled tasks (%s scheduled)..." %
            len(
                self.all_tasks))
        for task in self.all_tasks:

            now = datetime.datetime.utcnow()
            interval = datetime.timedelta(seconds=task["interval"])

            last_time = now - interval

            if task.get("dailytime"):

                dailytime = task.get("dailytime").time()

                time_datelastqueued = task.get(
                    "datelastqueued").time().isoformat()[0:8]
                time_dailytime = dailytime.isoformat()[0:8]
                if task.get(
                        "datelastqueued") and time_datelastqueued != time_dailytime:
                    log.debug(
                        "Adjusting the time of scheduled task %s from %s to %s" %
                        (task["_id"], time_datelastqueued, time_dailytime))

                    # Make sure we don't queue the task in a loop by adjusting
                    # the time
                    if time_datelastqueued < time_dailytime:
                        adjusted_datelastqueued = datetime.datetime.combine(
                            task.get("datelastqueued").date() -
                            datetime.timedelta(days=1),
                            dailytime)
                    else:
                        adjusted_datelastqueued = datetime.datetime.combine(
                            task.get("datelastqueued").date(), dailytime)

                    # We do find_and_modify and not update() because several check() may be happening
                    # at the same time.
                    self.collection.find_and_modify(
                        {
                            "_id": task["_id"],
                            "datelastqueued": task.get("datelastqueued")
                        },
                        {"$set": {
                            "datelastqueued": adjusted_datelastqueued
                        }}
                    )
                    self.refresh()

            task_data = self.collection.find_and_modify(
                {
                    "_id": task["_id"],
                    "datelastqueued": {"$lt": last_time}
                },
                {"$set": {
                    "datelastqueued": now
                }}
            )

            if task_data:
                send_task(
                    task_data["path"],
                    task_data["params"],
                    queue=task.get("queue"))
                log.debug("Scheduler: queued %s" % task_data)

                self.refresh()
